-- 20120302   mvh   Created
-- 20151015   mvh   WIP mouse wheel and more interface
-- 20151018   mvh   Cleanup, added default level and window; added error handling
-- 20151018   mvh   Note: print has a hardcoded limit at 2000 chars, avoid really long strings!
-- 20151019   mvh   Added keyboard interface
-- 20151020   mvh   Zoom and pan working
-- 20151021   mvh   Cleanup zoom and pan; show level/zoom etc info in table, frame control

-- default series information - this is here to allow debugging of the code usign ZeroBrane Studio, running from the server
local CGIseries = '09-759535:1.3.46.670589.11.0.0.11.4.2.0.5479.5.5112.2007011215371493050'
-- local CGIseries = '1082100772_3727835028:1.2.826.0.1.3680043.2.135.733552.44047153.7.1252971725.189.61'

------------------------------------------------------------------------
-- support functions 
------------------------------------------------------------------------

-- split string into pieces, return as table
function split(str, pat)
   local t = {} 
   local fpat = "(.-)" .. pat
   local last_end = 1
   local s, e, cap = str:find(fpat, 1)
   while s do
      if s ~= 1 or cap ~= "" then
	 table.insert(t,cap)
      end
      last_end = e+1
      s, e, cap = str:find(fpat, last_end)
   end
   if last_end <= #str then
      cap = str:sub(last_end)
      table.insert(t, cap)
   end
   return t
end

-- report error string on a otherwise empty reddish page
function errorpage(s)
  HTML('Content-type: text/html\n\n');
  print(
  [[
  <HEAD><TITLE>Version ]].. (version or '-')..[[ WADO viewer</TITLE></HEAD>
  <BODY BGCOLOR='FFDFCF'>
  <H3>Conquest ]].. (version or '-') ..[[ WADO viewer</H3>
  <H3>]].. s .. [[</H3>
  </BODY>
  ]])
end

-- get some extra information at series and study level
function getseriesinfo(source, patid, seriesuid)
  local b;
  b=newdicomobject();
  b.PatientID = patid
  b.SeriesInstanceUID = seriesuid
  b.StudyInstanceUID = '';
  b.Modality = '';
  b.SeriesDescription = '';
  b.SeriesDate = '';
  b.SeriesTime = '';
  a=dicomquery(source, 'SERIES', b);
  if a==nil then
   errorpage('Wado viewer error, cannot query server at series level: ' .. source)
   return
  end
  
  b=newdicomobject();
  b.PatientID = patid
  b.StudyInstanceUID = a[0].StudyInstanceUID;
  b.PatientName = '';
  b.PatientSex = '';
  b.PatientBirthDate = '';
  c=dicomquery(source, 'STUDY', b);
  if c==nil then
   errorpage('Wado viewer error, cannot query server at image level: ' .. source)
   return
  end
  
  return a[0], c[0]
end

-- get list of images
function queryimages(source, patid, seriesuid)
  local images,imaget,b;
  
  b=newdicomobject();
  b.PatientID        = patid
  b.SeriesInstanceUID= seriesuid
  b.InstanceNumber   = ''
  b.SOPInstanceUID   = ''
  b.SliceLocation   = ''
  b.ImageComments   = ''
  b.NumberOfFrames = ''

  images=dicomquery(source, 'IMAGE', b);
  if images==nil then
   errorpage('Wado viewer error, cannot query server at image level: ' .. source)
   return
  end

  imaget={}
  for k=0,#images-1 do
    if (images[k].InstanceNumber or '') == '' then
      images[k].InstanceNumber = '0'
    end
    imaget[k+1]={}
	  imaget[k+1].SOPInstanceUID=images[k].SOPInstanceUID
	  imaget[k+1].SliceLocation=images[k].SliceLocation or ''
	  imaget[k+1].ImageComments=images[k].ImageComments or ''
	  imaget[k+1].NumberOfFrames=images[k].NumberOfFrames or 1
	  imaget[k+1].InstanceNumber= tonumber(images[k].InstanceNumber or '0')
  end
  table.sort(imaget, function(a,b) return a.InstanceNumber<b.InstanceNumber end)

  return imaget
end

------------------------------------------------------------------------
-- Start main code
------------------------------------------------------------------------

-- get local server AE title from dicom.ini at server side
local source
if CGI('source', '(local)')=='(local)' then
  source = servercommand('get_param:MyACRNema')
  if source==nil then 
    errorpage('server not running: '.. CGI('source')) 
    return
  end
else
  source = CGI('source')
end

local bridge=''
if CGI('source', '(local)')~='(local)' then
	bridge='&bridge='..CGI('source')
end

-- collect all DICOM information needed
local patid, seriesuid = unpack(split(CGI('series', CGIseries), ':'))
local seriesinfo, patinfo = getseriesinfo(source, patid, seriesuid)
local studyuid = seriesinfo.StudyInstanceUID
local modality = seriesinfo.Modality
local description = seriesinfo.SeriesDescription
local date = seriesinfo.SeriesDate
local time = seriesinfo.SeriesTime
local images = queryimages(source, patid, seriesuid);

-- size of image display - comes from cgi-bin/dicom.ini
local size = 0+CGI('size', '512')

-- some error handling
if studyuid==nil then
  errorpage('wado viewer error: unknown series')
  return
end

if #images==0 then
  errorpage('wado viewer error: no images found')
  return
end

local a = newdicomobject()
a.InstanceNumber = 'test'
if a.InstanceNumber ~= 'test' then
  errorpage('wado viewer error: old dgate.dic used')
  return
end

---------------------------------------------------------------------------------
-- get some information out of an arbitrary slice at the server side
-- running a short lua script on the server
-- this is needed because not all information can be queried
-- and the CGI interface has no direct access to DICOM slices
---------------------------------------------------------------------------------

-- this notation is accepted in the server and will also access virtual servers
local imagelocation = studyuid..'\\\\'..seriesuid..'\\\\'..images[1].SOPInstanceUID

windowcenter, windowwidth, slope, intercept = 
  unpack(split(servercommand('lua:'..
  [[
  local a=newdicomobject(); 
  a:Read("]]..imagelocation..[[");
  local c = (a.WindowCenter or 1000)  if c=='' then c=1000 end
  local w = (a.WindowWidth or 1000)   if w=='' then w=1000 end
  local s = (a.RescaleSlope or 1)     if s=='' then c=1 end
  local i = (a.RescaleIntercept or 0) if i=='' then i=0 end
  print('getting image for wado viewer')
  return c.."|"..w.."|"..s.."|"..i
  ]]
  ), '|'))

-- process level and window information: may be multiple
if string.find(windowcenter, '%\\') then
  windowcenter = unpack(split(windowcenter, '\\'))
end

if string.find(windowwidth, '%\\') then
  windowwidth = unpack(split(windowwidth, '\\'))
end

-- todo: fully scale level and window according to slope and intercept
if modality=='CT' then
  windowcenter = windowcenter+1024 -- default for CT in conquest
end

---------------------------------------------------------------------------------
-- generate HTML page
---------------------------------------------------------------------------------

HTML('Content-type: text/html\n\n');
print([[
<HEAD><TITLE>Version ]].. (version or '-')..[[ WADO viewer</TITLE></HEAD>
<BODY BGCOLOR='CFDFCF'>
<H3>Conquest ]].. (version or '-') ..[[ WADO viewer</H3>

<SCRIPT language=JavaScript>
]])

-- bridge Lua to javascript with this block of variables 
print([[
var defwindowcenter = ]]..windowcenter..[[;
var defwindowwidth = ]]..windowwidth..[[;
var modality = ']]..modality..[[';
var description = ']]..description..[[';
var studyuid = ']]..studyuid..[[';
var seriesuid = ']]..seriesuid..[[';
var bridge = ']]..bridge..[[';
var size = ]]..size..[[;
var nframes = 1;
var region = '';
var zoom = 1;
var panx=0.5;
var pany=0.5;

var windowcenter = defwindowcenter;
var windowwidth = defwindowwidth;
var frame = 0;
var inter;
]])

-- javascript function to load the DICOM object as image or text
print([[
function load()
{ if (0+document.forms[0].slice.value.split("|")[1]!=nframes)
  { nframes = 0+document.forms[0].slice.value.split("|")[1];
    frame = 0;
  }

  if (modality.substring(0, 2)!='RT' || modality=='RTDOSE')
  { document.images[0].src = 'dgate.exe?requestType=WADO&contentType=image/jpeg'+
                           bridge +
                           '&studyUID='+studyuid +
                           '&seriesUID='+seriesuid +
                           '&windowCenter='+windowcenter +
                           '&windowWidth='+windowwidth +
                           '&frameNumber='+frame +
                           '&region='+region +
			   '&objectUID=' + document.forms[0].slice.value.split("|")[0];
			   
			   document.images[0].addEventListener("mousewheel", myWheelFunction);
			   document.addEventListener("keydown", myKeyFunction, false);
  }
  else  
    document.getElementById("myframe").src   = 'dgate.exe?requestType=WADO&contentType=text/plain'+
                           bridge+
                           '&studyUID='+studyuid +                           '&seriesUID='+seriesuid +
                           '&objectUID=' + document.forms[0].slice.value;
}
]])

-- javascript function for a browser popup (used to display header info)
print([[
function PopupCenter(pageURL, title,w,h)
{ var left = (screen.width/2)-(w/2);
  var top = (screen.height/2)-(h/2);
  var targetWin = 
  window.open (pageURL, title, 'toolbar=no, location=no, directories=no, status=no, menubar=no, scrollbars=yes, resizable=yes, copyhistory=no, width='+w+', height='+h+', top='+top+', left='+left);
}
]])

-- javascript function for wheel handling event
print([[
function myWheelFunction(a) 
{ a = a | event;
  if (a.altKey)
  { if (a.wheelDelta>0)
      nextslice(5)
    else
     prevslice(5);
    return;
  }

  if (a.wheelDelta>0)
    nextslice(1)
  else
   prevslice(1);
}
]])

print([[
function leveler(a, b)
{ if (b==0 && a==1)
  { windowwidth = defwindowwidth;
    windowcenter = defwindowcenter;
  }
  else if (b==0 && a==2)
  { windowwidth = 0;
    windowcenter = 0;
  }
  else
  { if (windowwidth==0)
    { windowwidth = defwindowwidth;
      windowcenter = defwindowcenter;
    }
    windowcenter = windowcenter + a*defwindowwidth;
    windowwidth = windowwidth * b;
  }
  load();
  document.getElementById("level").innerHTML = "level: "+Math.round(windowcenter*100)/100;
  document.getElementById("window").innerHTML = "window: "+Math.round(windowwidth*100)/100;
}

function framer(a)
{ frame = frame + a;
  if (frame>=nframes) frame = nframes-1;
  if (frame<0)        frame = 0;
  load();
  document.getElementById("frame").innerHTML = "frame: "+frame;
}
]])

print([[
function zoomer(z, px, py)
{ zoom = z*zoom;
  panx = panx + px/zoom;
  pany = pany + py/zoom;
  if (zoom<=1)
  { zoom=1;
    panx=0.5;
    pany=0.5;
    region="";
  }
  else
  { var xl, xh, yl, yh;
    if (zoom>50) zoom = 50;
    xl = panx - 0.5/zoom; if(xl<0)      {xl=0;     panx=0.5/zoom; }
    xh = panx + 0.5/zoom; if(xh>0.999)  {xh=0.999; panx=xh-0.5/zoom; }
    yl = pany - 0.5/zoom; if(yl<0)      {yl=0;     pany=0.5/zoom; }
    yh = pany + 0.5/zoom; if(yh>=0.999) {yh=0.999; pany=1-0.5/zoom; }
    region = String(xl)+","+String(xh)+","+String(yl)+","+String(yh);
  }
  load();
  document.getElementById("zoom").innerHTML = "zoom: "+Math.round(zoom*100)/100;
}
]])

-- javascript function for keyboard event
print([[
function myKeyFunction(a) 
{ a = event;
  var ch = String.fromCharCode(a.keyCode || a.charCode)
  if      (ch=='B') leveler(-0.05, 1);
  else if (ch=='D') leveler(+0.05, 1);
  else if (ch=='L') leveler(0, 1.2);
  else if (ch=='H') leveler(0, 1/1.2);
  else if (ch=='P') leveler(1, 0);
  else if (ch=='A') leveler(2, 0);
  else if (ch=='N') leveler(0, -1);

  else if (ch=='I') zoomer(1.2, 0, 0);
  else if (ch=='O') zoomer(1/1.2, 0, 0);
  else if (ch=='F') zoomer(0, 0, 0);
  else if (ch=='U') framer(1);
  else if (ch=='V') framer(-1);
  else if (a.keyCode==37) zoomer(1, 0.03, 0);
  else if (a.keyCode==39) zoomer(1, -0.03, 0);
  else if (a.keyCode==38) zoomer(1, 0, 0.03);
  else if (a.keyCode==40) zoomer(1, 0, -0.03);

  else if (a.keyCode==33) nextslice();
  else if (a.keyCode==34) prevslice();
  else if (ch=='H')
    PopupCenter('dgate.exe?requestType=WADO'+bridge+'&contentType=text/plain&studyUID='+studyuid+'&seriesUID='
    +seriesuid+'&objectUID=' + document.forms[0].slice.value, 'hoi', 700, 512);
}
]])

-- javascript functions to step through slices
print([[
function nextslice(delt)
{ delt = delt | 1;
  if (document.forms[0].slice.selectedIndex >= document.forms[0].slice.length-delt) 
    document.forms[0].slice.selectedIndex = document.forms[0].slice.length-1; 
  else 
    document.forms[0].slice.selectedIndex = document.forms[0].slice.selectedIndex + delt; 
  load()
}

function prevslice(delt)
{ delt = delt | 1;
  if (document.forms[0].slice.selectedIndex < delt) 
    document.forms[0].slice.selectedIndex = 0; 
  else document.forms[0].slice.selectedIndex = document.forms[0].slice.selectedIndex - delt; 
  load()
}

</SCRIPT>
]])

if time~='' then
  time = math.floor(time+0.5);
end

-- display info
print(string.format('<table style="width=560"><td>date %s</td><td>time %s</td>', date, time))
print('<td id=zoom>zoom: 1</td>')
print('<td id=level>level: 0</td>')
print('<td id=window>window: 0</td>')
print('<td id=frame>frame: 0</td>')
print('</table>')

-- generate element to display information: image (IMG) or text (iframe)
if string.sub(modality, 1, 2)=='RT' and modality~='RTDOSE' then
print([[
<iframe id="myframe" BORDER=0 WIDTH=]]..size..[[ HEIGHT=]]..size..[[>
<p>Your browser does not support iframes.</p>
</iframe>
]])
else
print([[
<IMG BORDER=0 HEIGHT=]]..size..[[>
]])
end

-- generate text string with patient and series information
print([[
<H4>]].. string.format('%s (ID %s; gender %s; born %s) %s %s', patinfo.PatientName, patinfo.PatientID, patinfo.PatientSex, patinfo.PatientBirthDate,  modality, description) ..[[</H4>
]])

-- generate form with slicing controls
print([[
<FORM>
Slice:
  <INPUT TYPE=BUTTON VALUE='<' onclick=prevslice() onmousedown="inter=setInterval(prevslice, 100)" onmouseup=clearInterval(inter) onmouseout=clearInterval(inter)>
  <select name=slice onchange=load() >
]])

for i=1, math.min(#images, 1000) do
  print('<option value='..images[i].SOPInstanceUID..'|'..images[i].NumberOfFrames..'>'..string.format('%d (%.1f) %s', i, tonumber(images[i].SliceLocation) or 0, images[i].ImageComments or '') .. '</option>')
end

if #images>1000 then
  print('<option value='..images[#images].SOPInstanceUID..'> list truncated </option>')
end

print([[
  </select>
  <INPUT TYPE=BUTTON VALUE='>' onclick=nextslice() onmousedown="inter=setInterval(nextslice, 100)" onmouseup=clearInterval(inter) onmouseout=clearInterval(inter)> Total slices: ]].. #images .. [[
</FORM>
]])

-- generate level and window preset choices and shorcut to show header
print([[
<br>
<a href=# onclick="windowcenter=defwindowcenter; windowwidth=defwindowwidth; load()">[lw preset default]</a>
<a href=# onclick="windowcenter=1000; windowwidth=1000; load()">[lw preset 1]</a>
<a href=# onclick="windowcenter=1000; windowwidth=400; load()">[lw preset 2]</a>
<a href=# onclick="windowcenter=''; windowwidth=''; load()">[lw preset auto]</a>

<br>
<a href=# onclick="javascript:PopupCenter('dgate.exe?requestType=WADO'+bridge+'&contentType=text/plain&studyUID='+studyuid+'&seriesUID='+seriesuid+'&objectUID=' + document.forms[0].slice.value, 'hoi', 700, 512)">[show header]</a>
]])

-- generate page close and autoload of first image
print([[
<SCRIPT language=JavaScript>
  document.onload=document.forms[0].slice.selectedIndex= document.forms[0].slice.length/2;leveler(1,0)
</SCRIPT>
</BODY>
]])
